from setuptools import setup

setup(name='semiolog',
      version='0.0.0',
      description='Semiological analysis',
      url='https://gitlab.ethz.ch/semiomaths/semiolog',
      author='Gianni Gastaldi',
      author_email='giannigastaldi@gmail.com',
      license='CC BY-NC 4.0',
      packages=['semiolog'],
      zip_safe=False)